package uk.co.thebadgerset.junit.maven;

import java.io.File;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.*;

import junit.textui.TestRunner;

import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;

import uk.co.thebadgerset.junit.extensions.TKTestRunner;

/**
 * <p><table id="crc"><caption>CRC Card</caption>
 * <tr><th> Responsibilities <th> Collaborations
 * </table>
 *
 * @author Rupert Smith
 *
 * @goal tktest
 * @phase test
 * @execute phase="test"
 * @requiresDependencyResolution test
 */
public class TKTestRunnerMojo extends AbstractMojo
{
    private static final BitSet UNRESERVED = new BitSet(256);

    /**
     * The TKTest runner command lines. There are passed directly to the TKTestRunner main method.
     *
     * @parameter
     */
    private Map<String, String> commands = new LinkedHashMap<String, String>();

    /**
     * The directory containing generated classes of the project being tested.
     *
     * @parameter expression="${project.build.outputDirectory}"
     * @required
     */
    private File classesDirectory;

    /**
     * The directory containing generated test classes of the project being tested.
     *
     * @parameter expression="${project.build.testOutputDirectory}"
     * @required
     */
    private File testClassesDirectory;

    /**
     * The classpath elements of the project being tested.
     *
     * @parameter expression="${project.testClasspathElements}"
     * @required
     * @readonly
     */
    private List classpathElements;

    /**
     * List of System properties to pass to the tests.
     *
     * @parameter
     */
    private Properties systemProperties;

    /**
     * Map of of plugin artifacts.
     *
     * @parameter expression="${plugin.artifactMap}"
     * @required
     * @readonly
     */
    private Map pluginArtifactMap;

    /**
     * Map of of project artifacts.
     *
     * @parameter expression="${project.artifactMap}"
     * @required
     * @readonly
     */
    private Map projectArtifactMap;

    /**
     * Option to specify the forking mode. Can be "never" (default), "once" or "always".
     * "none" and "pertest" are also accepted for backwards compatibility.
     *
     * @parameter expression="${forkMode}" default-value="once"
     */
    private String forkMode;

    /**
     * Option to specify the jvm (or path to the java executable) to use with
     * the forking options. For the default we will assume that java is in the path.
     *
     * @parameter expression="${jvm}"
     * default-value="java"
     */
    private String jvm;

    public static URL getURL(File file) throws MalformedURLException
    {
        URL url = file.toURL();
        // encode any characters that do not comply with RFC 2396
        // this is primarily to handle Windows where the user's home directory contains spaces
        /*try
        {
            byte[] bytes = url.toString().getBytes("US-ASCII");
            StringBuffer buf = new StringBuffer(bytes.length);
            for (int i = 0; i < bytes.length; i++)
            {
                byte b = bytes[i];
                if (UNRESERVED.get(b))
                {
                    buf.append((char) b);
                }
                else
                {
                    buf.append('%');
                    buf.append(Character.forDigit((b >>> 4) & 0xf, 16));
                    buf.append(Character.forDigit(b & 0xf, 16));
                }
            }
        
            return new URL(buf.toString());
        }
        catch (UnsupportedEncodingException e)
        {
            // should not happen as US-ASCII must be present
            throw new RuntimeException(e);
        }*/

        return url;
    }

    private static ClassLoader createClassLoader(List classPathUrls, ClassLoader parent, boolean childDelegation)
                                          throws MalformedURLException
    {
        List urls = new ArrayList();

        for (Iterator i = classPathUrls.iterator(); i.hasNext();)
        {
            String url = (String) i.next();

            if (url != null)
            {
                File f = new File(url);
                urls.add(getURL(f));
            }
        }

        IsolatedClassLoader classLoader = new IsolatedClassLoader(parent, childDelegation);

        for (Iterator iter = urls.iterator(); iter.hasNext();)
        {
            URL url = (URL) iter.next();
            classLoader.addURL(url);
        }

        return classLoader;
    }

    /**
     * Implementation of the tktest goal.
     *
     * @throws MojoExecutionException
     */
    public void execute() throws MojoExecutionException
    {
        getLog().info("Test Classpath :");

        // no need to add classes/test classes directory here - they are in the classpath elements already
        for (Iterator i = classpathElements.iterator(); i.hasNext();)
        {
            String classpathElement = (String) i.next();
            getLog().info("  " + classpathElement);
        }

        try
        {
            //ClassLoader oldContextClassLoader = Thread.currentThread().getContextClassLoader();
            ClassLoader testsClassLoader = ClassLoader.getSystemClassLoader();

            ClassLoader tkRunnerClassLoader = createClassLoader(classpathElements, testsClassLoader, true);

            Class tkRunnerClass = tkRunnerClassLoader.loadClass(TKTestRunner.class.getName());

            Object surefire = tkRunnerClass.newInstance();

            Method run = tkRunnerClass.getMethod("main", new Class[] { String[].class });

            Thread.currentThread().setContextClassLoader(testsClassLoader);

            // Pass each of the test runner command lines in turn to the toolkit test runner.
            for (String testName : commands.keySet())
            {
                String commandLine = commands.get(testName);

                // Tokenize the command line on white space, into an array of string components.
                String[] tokenizedCommandLine = commandLine.split("\\s");

                // Run the tests.
                //OldTKTestRunner.main(tokenizedCommandLine);
                run.invoke(surefire, new Object[] { tokenizedCommandLine });
            }
        }
        catch (Exception e)
        {
            getLog().error("There was an exception: " + e.getMessage(), e);
        }
    }
}
